from pytorch_widedeep.models.wide import Wide
from pytorch_widedeep.models.tab_mlp import TabMlp
from pytorch_widedeep.models.deep_text import DeepText
from pytorch_widedeep.models.wide_deep import WideDeep
from pytorch_widedeep.models.deep_image import DeepImage
from pytorch_widedeep.models.tab_resnet import TabResnet
from pytorch_widedeep.models.tab_transformer import TabTransformer
